/*
 * Copyright 2011 Jacek Caban for CodeWeavers
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 */

#define DISPID_SUBMATCHES_COUNT     1

#define DISPID_MATCHCOLLECTION_COUNT    1

#define DISPID_MATCH_FIRSTINDEX     10001
#define DISPID_MATCH_LENGTH         10002
#define DISPID_MATCH_SUBMATCHES     10003

#define DISPID_REGEXP_PATTERN       10001
#define DISPID_REGEXP_IGNORECASE    10002
#define DISPID_REGEXP_GLOBAL        10003
#define DISPID_REGEXP_EXECUTE       10004
#define DISPID_REGEXP_TEST          10005
#define DISPID_REGEXP_REPLACE       10006
#define DISPID_REGEXP_MULTILINE     10007
