/*
 *{{NO_DEPENDENCIES}}
 * Microsoft Developer Studio generated include file.
 * Used by taskmgr.rc
 *
 */

#include <windef.h>
#include <winuser.h>
#include <commctrl.h>

#define IDS_APP_TITLE                   103
#define IDC_TASKMGR                     109
#define IDS_RUNDLG_CAPTION              110

#define IDD_TASKMGR_DIALOG              102
#define IDM_ABOUT                       104
#define IDM_EXIT                        105
#define IDD_APPLICATION_PAGE            106
#define IDI_TASKMGR                     107
#define IDI_SMALL                       108
#define IDR_MAINFRAME                   128
#define IDR_TASKMANAGER                 130
#define IDD_PROCESS_PAGE                133
#define IDD_PERFORMANCE_PAGE            134
#define IDR_WINDOWSMENU                 135
#define IDI_TASKMANAGER                 136
#define IDI_WINDOW                      137
#define IDI_WINDOWSM                    138
#define IDR_APPLICATION_PAGE_CONTEXT1   139
#define IDR_APPLICATION_PAGE_CONTEXT2   140
#define IDR_TRAY_POPUP                  141
#define IDD_AFFINITY_DIALOG             142
#define IDD_COLUMNS_DIALOG              143
#define IDR_PROCESS_PAGE_CONTEXT        144
#define IDB_TRAYMASK                    150
#define IDB_TRAYICON                    153
#define IDB_FONT                        154
#define IDD_DEBUG_CHANNELS_DIALOG       155
#define IDC_DEBUG_CHANNELS_LIST         156

#define IDC_ENDTASK                     1012
#define IDC_SWITCHTO                    1013
#define IDC_NEWTASK                     1014
#define IDC_TAB                         1015
#define IDC_APPLIST                     1016
#define IDC_ENDPROCESS                  1017
#define IDC_PROCESSLIST                 1018
#define IDC_CPU0                        1019
#define IDC_CPU1                        1020
#define IDC_CPU2                        1021
#define IDC_SHOWALLPROCESSES            1021
#define IDC_CPU3                        1022
#define IDC_IMAGENAME                   1022
#define IDC_CPU4                        1023
#define IDC_PID                         1023
#define IDC_CPU5                        1024
#define IDC_CPUUSAGE                    1024
#define IDC_TOTALS_HANDLE_COUNT         1024
#define IDC_CPU6                        1025
#define IDC_CPUTIME                     1025
#define IDC_CPU7                        1026
#define IDC_MEMORYUSAGE                 1026
#define IDC_TOTALS_THREAD_COUNT         1026
#define IDC_CPU8                        1027
#define IDC_MEMORYUSAGEDELTA            1027
#define IDC_TOTALS_PROCESS_COUNT        1027
#define IDC_CPU9                        1028
#define IDC_PEAKMEMORYUSAGE             1028
#define IDC_COMMIT_CHARGE_TOTAL         1028
#define IDC_CPU10                       1029
#define IDC_PAGEFAULTS                  1029
#define IDC_COMMIT_CHARGE_LIMIT         1029
#define IDC_CPU11                       1030
#define IDC_USEROBJECTS                 1030
#define IDC_COMMIT_CHARGE_PEAK          1030
#define IDC_CPU12                       1031
#define IDC_IOREADS                     1031
#define IDC_PHYSICAL_MEMORY_TOTAL       1031
#define IDC_CPU13                       1032
#define IDC_IOREADBYTES                 1032
#define IDC_PHYSICAL_MEMORY_AVAILABLE   1032
#define IDC_CPU14                       1033
#define IDC_SESSIONID                   1033
#define IDC_PHYSICAL_MEMORY_SYSTEM_CACHE 1033
#define IDC_CPU15                       1034
#define IDC_USERNAME                    1034
#define IDC_KERNEL_MEMORY_TOTAL         1034
#define IDC_CPU16                       1035
#define IDC_PAGEFAULTSDELTA             1035
#define IDC_KERNEL_MEMORY_PAGED         1035
#define IDC_CPU17                       1036
#define IDC_VIRTUALMEMORYSIZE           1036
#define IDC_KERNEL_MEMORY_NONPAGED      1036
#define IDC_CPU18                       1037
#define IDC_PAGEDPOOL                   1037
#define IDC_TOTALS_FRAME                1037
#define IDC_CPU19                       1038
#define IDC_NONPAGEDPOOL                1038
#define IDC_COMMIT_CHARGE_FRAME         1038
#define IDC_CPU20                       1039
#define IDC_BASEPRIORITY                1039
#define IDC_KERNEL_MEMORY_FRAME         1039
#define IDC_CPU21                       1040
#define IDC_HANDLECOUNT                 1040
#define IDC_PHYSICAL_MEMORY_FRAME       1040
#define IDC_CPU22                       1041
#define IDC_THREADCOUNT                 1041
#define IDC_CPU23                       1042
#define IDC_CPU24                       1043
#define IDC_GDIOBJECTS                  1043
#define IDC_CPU_USAGE_FRAME             1043
#define IDC_CPU25                       1044
#define IDC_IOWRITES                    1044
#define IDC_MEM_USAGE_FRAME             1044
#define IDC_CPU26                       1045
#define IDC_IOWRITEBYTES                1045
#define IDC_CPU_USAGE_HISTORY_FRAME     1045
#define IDC_CPU27                       1046
#define IDC_IOOTHER                     1046
#define IDC_MEMORY_USAGE_HISTORY_FRAME  1046
#define IDC_CPU28                       1047
#define IDC_IOOTHERBYTES                1047
#define IDC_CPU_USAGE_GRAPH             1047
#define IDC_CPU29                       1048
#define IDC_MEM_USAGE_GRAPH2            1048
#define IDC_MEM_USAGE_GRAPH             1048
#define IDC_CPU30                       1049
#define IDC_MEM_USAGE_HISTORY_GRAPH     1049
#define IDC_CPU_USAGE_HISTORY_GRAPH     1050
#define IDC_CPU31                       1051

#define IDS_TOTALS_HANDLE_COUNT         1060
#define IDS_TOTALS_THREAD_COUNT         1061
#define IDS_TOTALS_PROCESS_COUNT        1062
#define IDS_COMMIT_CHARGE_TOTAL         1063
#define IDS_COMMIT_CHARGE_LIMIT         1064
#define IDS_COMMIT_CHARGE_PEAK          1065
#define IDS_PHYSICAL_MEMORY_TOTAL       1066
#define IDS_PHYSICAL_MEMORY_AVAILABLE   1067
#define IDS_PHYSICAL_MEMORY_SYSTEM_CACHE 1068
#define IDS_KERNEL_MEMORY_TOTAL         1069
#define IDS_KERNEL_MEMORY_PAGED         1070
#define IDS_KERNEL_MEMORY_NONPAGED      1071


#define ID_FILE_NEW                     32771
#define ID_OPTIONS_ALWAYSONTOP          32773
#define ID_OPTIONS_MINIMIZEONUSE        32774
#define ID_OPTIONS_HIDEWHENMINIMIZED    32775
#define ID_VIEW_REFRESH                 32776
#define ID_VIEW_LARGE                   32778
#define ID_VIEW_SMALL                   32779
#define ID_VIEW_DETAILS                 32780
#define ID_VIEW_UPDATESPEED_HIGH        32781
#define ID_VIEW_UPDATESPEED_NORMAL      32782
#define ID_VIEW_UPDATESPEED_LOW         32783
#define ID_VIEW_UPDATESPEED_PAUSED      32784
#define ID_WINDOWS_TILEHORIZONTALLY     32785
#define ID_WINDOWS_TILEVERTICALLY       32786
#define ID_WINDOWS_MINIMIZE             32787
#define ID_WINDOWS_MAXIMIZE             32788
#define ID_WINDOWS_CASCADE              32789
#define ID_WINDOWS_BRINGTOFRONT         32790
#define ID_HELP_TOPICS                  32791
#define ID_HELP_ABOUT                   32792
#define ID_FILE_EXIT                    32793
#define ID_OPTIONS_SHOW16BITTASKS       32794
#define ID_VIEW_SELECTCOLUMNS           32795
#define ID_VIEW_SHOWKERNELTIMES         32796
#define ID_VIEW_CPUHISTORY_ONEGRAPHALL  32797
#define ID_VIEW_CPUHISTORY_ONEGRAPHPERCPU 32798
#define ID_APPLICATION_PAGE_SWITCHTO    32799
#define ID_ENDTASK                      32800
#define ID_GOTOPROCESS                  32801
#define ID_RESTORE                      32802
#define ID_APPLICATION_PAGE_ENDTASK     32803
#define ID_APPLICATION_PAGE_GOTOPROCESS 32804
#define ID_PROCESS_PAGE_ENDPROCESS      32805
#define ID_PROCESS_PAGE_ENDPROCESSTREE  32806
#define ID_PROCESS_PAGE_DEBUG           32807
#define ID_PROCESS_PAGE_SETAFFINITY     32808
#define ID_PROCESS_PAGE_SETPRIORITY_REALTIME 32809
#define ID_PROCESS_PAGE_SETPRIORITY_HIGH 32810
#define ID_PROCESS_PAGE_SETPRIORITY_ABOVENORMAL 32811
#define ID_PROCESS_PAGE_SETPRIORITY_NORMAL 32812
#define ID_PROCESS_PAGE_SETPRIORITY_BELOWNORMAL 32813
#define ID_PROCESS_PAGE_SETPRIORITY_LOW 32814
#define ID_PROCESS_PAGE_DEBUGCHANNELS   32815
#define IDS_VIEW_LARGE                  32816
#define IDS_VIEW_SMALL                  32817
#define IDS_VIEW_DETAILS                32818
#define IDS_WINDOWS                     32819
#define IDS_VIEW_SELECTCOLUMNS          32820
#define IDS_OPTIONS_SHOW16BITTASKS      32821
#define IDS_VIEW_CPUHISTORY             32822
#define IDS_VIEW_CPUHISTORY_ONEGRAPHALL 32823
#define IDS_VIEW_CPUHISTORY_ONEGRAPHPERCPU 32824
#define IDS_VIEW_SHOWKERNELTIMES        32825
#define IDS_APPLICATIONS                32826
#define IDS_PROCESSES                   32827
#define IDS_PERFORMANCE                 32828
#define IDS_STATUS_BAR_CPU_USAGE        32829
#define IDS_STATUS_BAR_PROCESSES        32830
#define IDS_STATUS_BAR_MEMORY_USAGE     32831
#define IDS_IMAGENAME                   32832
#define IDS_PID                         32833
#define IDS_CPUUSAGE                    32834
#define IDS_CPUTIME                     32835
#define IDS_MEMORYUSAGE                 32836
#define IDS_MEMORYUSAGEDELTA            32837
#define IDS_PEAKMEMORYUSAGE             32838
#define IDS_PAGEFAULTS                  32839
#define IDS_USEROBJECTS                 32840
#define IDS_IOREADS                     32841
#define IDS_IOREADBYTES                 32842
#define IDS_SESSIONID                   32843
#define IDS_USERNAME                    32844
#define IDS_PAGEFAULTSDELTA             32845
#define IDS_VIRTUALMEMORYSIZE           32846
#define IDS_PAGEDPOOL                   32847
#define IDS_NONPAGEDPOOL                32848
#define IDS_BASEPRIORITY                32849
#define IDS_HANDLECOUNT                 32850
#define IDS_THREADCOUNT                 32851
#define IDS_GDIOBJECTS                  32852
#define IDS_IOWRITES                    32853
#define IDS_IOWRITEBYTES                32854
#define IDS_IOOTHER                     32855
#define IDS_IOOTHERBYTES                32856
#define IDS_WARNING_TITLE               32857
#define IDS_PRIORITY_CHANGE_MESSAGE     32858
#define IDS_PRIORITY_UNABLE2CHANGE      32859
#define IDS_TERMINATE_MESSAGE           32860
#define IDS_TERMINATE_UNABLE2TERMINATE  32861
#define IDS_DEBUG_MESSAGE               32862
#define IDS_DEBUG_UNABLE2DEBUG          32863
#define IDS_AFFINITY_ERROR_MESSAGE      32864
#define IDS_AFFINITY_ERROR_TITLE        32865
#define IDS_AFFINITY_UNABLE2ACCESS      32866
#define IDS_SYSTEM_IDLE_PROCESS         32867
#define IDS_APPLICATION_NOT_RESPONDING  32868
#define IDS_APPLICATION_RUNNING         32869
#define IDS_APPLICATION_TASK            32870
#define IDS_APPLICATION_STATUS          32871
#define IDS_DEBUG_CHANNEL               32872
#define IDC_STATIC                      -1

/*
 * Next default values for new objects
 *
 */
#ifdef APSTUDIO_INVOKED
#ifndef APSTUDIO_READONLY_SYMBOLS
#define _APS_NEXT_RESOURCE_VALUE        157
#define _APS_NEXT_COMMAND_VALUE         32817
#define _APS_NEXT_CONTROL_VALUE         1048
#define _APS_NEXT_SYMED_VALUE           110
#endif
#endif
