/*
 * Copyright 2007 Tim Schwartz
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 */

#include <windef.h>

#define STRING_USAGE              101
#define STRING_START_USAGE        102
#define STRING_STOP_USAGE         103
#define STRING_STOP_DEP           104
#define STRING_CANT_STOP          105
#define STRING_NO_SCM             106
#define STRING_NO_SVCHANDLE       107
#define STRING_START_SVC          108
#define STRING_START_SVC_SUCCESS  109
#define STRING_START_SVC_FAIL     110
#define STRING_STOP_SVC           111
#define STRING_STOP_SVC_SUCCESS   112
#define STRING_STOP_SVC_FAIL      113
#define STRING_NO_ENTRIES	  115
#define STRING_USE_HEADER	  116
#define STRING_USE_ENTRY	  117
#define STRING_OK		  118
#define STRING_PAUSED		  119
#define STRING_SESSLOST 	  120
#define STRING_NETERR		  121
#define STRING_CONN		  122
#define STRING_RECONN		  123
#define STRING_RUNNING_HEADER	  124
