/*
 * Copyright 2014 Piotr Caban for CodeWeavers
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 */

#define IDS_STATE_NORMAL            0x1000
#define IDS_STATE_UNAVAILABLE       0x1001
#define IDS_STATE_SELECTED          0x1002
#define IDS_STATE_FOCUSED           0x1003
#define IDS_STATE_PRESSED           0x1004
#define IDS_STATE_CHECKED           0x1005
#define IDS_STATE_MIXED             0x1006
#define IDS_STATE_READONLY          0x1007
#define IDS_STATE_HOTTRACKED        0x1008
#define IDS_STATE_DEFAULT           0x1009
#define IDS_STATE_EXPANDED          0x100a
#define IDS_STATE_COLLAPSED         0x100b
#define IDS_STATE_BUSY              0x100c
#define IDS_STATE_FLOATING          0x100d
#define IDS_STATE_MARQUEED          0x100e
#define IDS_STATE_ANIMATED          0x100f
#define IDS_STATE_INVISIBLE         0x1010
#define IDS_STATE_OFFSCREEN         0x1011
#define IDS_STATE_SIZEABLE          0x1012
#define IDS_STATE_MOVEABLE          0x1013
#define IDS_STATE_SELFVOICING       0x1014
#define IDS_STATE_FOCUSABLE         0x1015
#define IDS_STATE_SELECTABLE        0x1016
#define IDS_STATE_LINKED            0x1017
#define IDS_STATE_TRAVERSED         0x1018
#define IDS_STATE_MULTISELECTABLE   0x1019
#define IDS_STATE_EXTSELECTABLE     0x101a
#define IDS_STATE_ALERT_LOW         0x101b
#define IDS_STATE_ALERT_MEDIUM      0x101c
#define IDS_STATE_ALERT_HIGH        0x101d
#define IDS_STATE_PROTECTED         0x101e
#define IDS_STATE_HASPOPUP          0x101f
