/*
 * Copyright 2007 Google (Dan Hipschman)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 */

/* The header file generated from this includes rpc.h, which ends up
   including oaidl.h at some point.  This will cause all sorts of errors
   so the easiest thing to do is just comment out our entire header.  */

#pragma makedep proxy
#pragma makedep register

cpp_quote("#if 0    /* oleaut32_oaidl.idl hack */")
#include "oaidl.idl"
cpp_quote("#endif   /* oleaut32_oaidl.idl hack */")

[
    threading(both),
    uuid(b196b286-bab4-101a-b69c-00aa00341d07)
]
coclass PSFactoryBuffer { interface IFactoryBuffer; }
