/*
 * Copyright (C) 2005 Vijay Kiran Kamuju
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 */

import "objidl.idl";

[
  local,
  object,
  uuid(f77459a0-bf9a-11cf-ba4e-00c04fd70816),
  pointer_default(unique)
]
interface IMimeInfo : IUnknown
{
    typedef [unique] IMimeInfo *LPMIMEINFO;

    HRESULT GetMimeCLSIDMapping( [out] UINT      *pcTypes,
                                 [out] LPCSTR  * *ppszTypes,
                                 [out] CLSID   * *ppclsID );
}

cpp_quote("#define SID_IMimeInfo IID_IMimeInfo")
